package engine

import (
	"fmt"
	"github.com/dop251/goja"
	"github.com/dop251/goja_nodejs/require"
	"os"
)

type Engine struct {
	vm            *goja.Runtime
	cfg           Config
	registry      *require.Registry
	requireModule *require.RequireModule
}

type Config interface {
	IsStrictMode() bool
	WillPrintResults() bool
	GlobalModulePath() []string
	RuntimeWarning(err error)
}

func NewEngine(cfg Config) *Engine {
	reg := require.NewRegistry(require.WithGlobalFolders(cfg.GlobalModulePath()...))
	e := &Engine{
		vm:       goja.New(),
		cfg:      cfg,
		registry: reg,
	}
	return e
}

func (e *Engine) Prepare() error {
	e.requireModule = e.registry.Enable(e.vm)
	return nil
}

func (e *Engine) RunFile(file string) error {
	fjs, err := os.ReadFile(file)
	if err != nil {
		return fmt.Errorf("failed to read script file:\n%w", err)
	}
	compiled, err := goja.Compile(file, string(fjs), e.cfg.IsStrictMode())
	res, err := e.vm.RunProgram(compiled)
	if err != nil {
		return fmt.Errorf("error in script:\n%w", err)
	}
	if e.cfg.WillPrintResults() {
		fmt.Println(res.String())
	}
	return nil
}

func (e *Engine) AddGlobal(m IModule) {
	rf := m.Enable(&ModuleContext{
		engine: e,
	})
	rf(func(name string, value interface{}) {
		err := e.vm.Set(name, value)
		if err != nil {
			e.cfg.RuntimeWarning(fmt.Errorf("failed to add '%s' to global object: %w", name, err))
		}
	})
}

func (e *Engine) AddModule(modName string, m IModule) {
	rf := m.Enable(&ModuleContext{
		engine: e,
	})
	e.registry.RegisterNativeModule(modName, func(runtime *goja.Runtime, module *goja.Object) {
		exports := module.Get("exports").(*goja.Object)
		rf(func(name string, value interface{}) {
			err := exports.Set(name, value)
			if err != nil {
				e.cfg.RuntimeWarning(fmt.Errorf("failed to add '%s' to module '%s': %w", name, modName, err))
			}
		})
	})
}
